#   Copyright 2018 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
OS MQTT Connectivity Service module.

Contains OSMQTTConnectivityService class used for MQTT connection.
"""

from paho.mqtt import client as mqtt

from wolk.wolkcore import ConnectivityService
from wolk.wolkcore import InboundMessage
from wolk import LoggerFactory


class OSMQTTConnectivityService(ConnectivityService):
    """This class handles sending and receiving MQTT messages.

    :ivar ca_cert: Path to ca.crt file used for TLS
    :vartype ca_cert: str
    :ivar client: MQTT client used to send/receive data
    :vartype client: paho.mqtt.Client
    :ivar connected: State of the connection
    :vartype connected: bool
    :ivar connected_rc: return code of the connection
    :vartype connected_rc: int
    :ivar device: Holds authentication information and actuator references
    :vartype device: wolk.Device
    :ivar host: Address of the MQTT broker
    :vartype host: str
    :ivar inbound_message_listener: Callback method for inbound messages
    :vartype inbound_message_listener: method
    :ivar logger: Logger instance issued from the LoggerFactory class
    :vartype logger: logger
    :ivar port: Port to which to send messages
    :vartype port: int
    :ivar qos: Quality of Service for the MQTT connection
    :vartype qos: int
    :ivar topics: List of topics to subscribe to
    :vartype topics: list
    """

    def __init__(
        self,
        device,
        qos=2,
        host="api-demo.wolkabout.com",
        port=8883,
        ca_cert=None,
    ):
        """Provide the connection to the WolkAbout IoT Platform.

        :param device: Contains device key, password and actuator references
        :type device: Device
        :param qos: Quality of Service for MQTT connection, defaults to 2
        :type qos: int, optional
        :param host: Address of the MQTT broker, defaults to the Demo instance
        :type host: str, optional
        :param port: Port to which to send messages, defaults to 8883
        :type port: int, optional
        :param ca_cert: The certificate file used to encrypt the connection
        :type ca_cert: None, optional
        """
        self.device = device
        self.qos = qos
        self.host = host
        self.port = port
        self.connected = False
        self.connected_rc = None
        self.inbound_message_listener = None
        self.client = None
        self.topics = None
        self.ca_cert = ca_cert
        self.logger = LoggerFactory.logger_factory.get_logger(
            str(self.__class__.__name__)
        )
        self.logger.debug(
            "Init:  Device key: %s ; Device password: %s ;"
            " Actuator references: %s ;"
            " QoS: %s ; Host: %s ; Port: %s ; CA certificate: %s ",
            self.device.key,
            self.device.password,
            self.device.actuator_references,
            self.qos,
            self.host,
            self.port,
            self.ca_cert,
        )

    def set_inbound_message_listener(self, on_inbound_message):
        """
        Set the callback method to handle inbound messages.

        :param on_inbound_message: The method that handles inbound messages
        :type on_inbound_message: method
        """
        self.logger.debug(
            "set_inbound_message_listener called - "
            "inbound_message_listener set to: %s",
            on_inbound_message,
        )
        self.inbound_message_listener = on_inbound_message

    def on_mqtt_message(self, client, userdata, message):
        """
        Serialize inbound messages and passes them to inbound message listener.

        :param client: Client that received the message
        :type client: paho.mqtt.Client
        :param userdata: private user data set in Client()
        :type userdata: str
        :param message: Class with members: topic, payload, qos, retain.
        :type message: paho.mqtt.MQTTMessage

        :returns: Nothing
        :rtype: None if no message is present
        """
        self.logger.debug("on_mqtt_message called")
        if not message:
            return

        channel = message.topic

        payload = (
            message.payload
            if message.topic.startswith("service/binary")
            else message.payload.decode("utf-8")
        )

        message = InboundMessage(channel, payload)
        self.inbound_message_listener(message)
        if isinstance(message.payload, bytes):
            self.logger.debug(
                "Received from channel: %s ; Payload size: %s",
                message.channel,
                len(message.payload),
            )
        else:
            self.logger.debug(
                "Received from channel: %s ; Payload: %s",
                message.channel,
                message.payload,
            )

    def on_mqtt_connect(self, client, userdata, flags, rc):
        """
        Handle when the client receives a CONNACK response from the server.

        :param client: Client that received the message
        :type client: paho.mqtt.Client
        :param userdata: private user data set in Client()
        :type userdata: str
        :param flags: response flags sent by the broker
        :type flags: int
        :param rc: the connection result
        :type rc: int
        """
        self.logger.debug("on_mqtt_connect called with rc: %s", rc)
        if rc == 0:  # Connection successful

            self.connected = True
            self.connected_rc = 0
            # Subscribing in on_mqtt_connect() means if we lose the connection
            # and reconnect then subscriptions will be renewed.
            if self.topics:
                self.client.subscribe(self.topics)
            self.logger.debug(
                "on_mqtt_connect - self.connected : %s", self.connected
            )
        elif rc == 1:  # Connection refused - incorrect protocol version
            self.connected_rc = 1
        elif rc == 2:  # Connection refused - invalid client identifier
            self.connected_rc = 2
        elif rc == 3:  # Connection refused - server unavailable
            self.connected_rc = 3
        elif rc == 4:  # Connection refused - bad username or password
            self.connected_rc = 4
        elif rc == 5:  # Connection refused - not authorised
            self.connected_rc = 5

    def on_mqtt_disconnect(self, client, userdata, rc):
        """
        Handle when the client disconnects from the broker.

        :param client: Client that received the message
        :type client: paho.mqtt.Client
        :param userdata: private user data set in Client()
        :type userdata: str
        :param rc: the disconnection result
        :type rc: int

        :raises RuntimeError: Unexpected disconnection
        """
        if rc != 0:
            raise RuntimeError("Unexpected disconnection.")
        self.connected = False
        self.connected_rc = None
        self.logger.debug(
            "on_mqtt_disconnect - self.connected : %s", self.connected
        )

    def connect(self):
        """
        Establish the connection to the WolkAbout IoT platform.

        If there are actuators it will subscribe to topics
        that will contain actuator commands.
        Subscribes to firmware update related topics.
        Starts a loop to handle inbound messages.

        :returns: if self.connected returns None
        :rtype: None

        :raises RuntimeError: Reason for connection being refused
        """
        if self.connected:

            return

        self.logger.debug("connect started")
        self.client = mqtt.Client(
            client_id=self.device.key, clean_session=True
        )
        self.client.on_connect = self.on_mqtt_connect
        self.client.on_disconnect = self.on_mqtt_disconnect
        self.client.on_message = self.on_mqtt_message
        if self.ca_cert:
            self.client.tls_set(self.ca_cert)
            self.client.tls_insecure_set(True)
        self.client.username_pw_set(self.device.key, self.device.password)
        self.client.will_set(
            "lastwill/" + self.device.key, "Gone offline", 2, False
        )

        self.logger.debug(
            "client.connect called : host: %s ; port:%s, ca_cert:%s "
            "; username:%s ; password:%s",
            self.host,
            self.port,
            self.ca_cert,
            self.device.key,
            self.device.password,
        )
        self.client.connect(self.host, self.port)

        self.client.loop_start()

        while True:

            if self.connected_rc is None:
                continue

            if self.connected_rc == 0:
                break

            elif self.connected_rc == 1:
                raise RuntimeError(
                    "Connection refused - incorrect protocol version"
                )
                break

            elif self.connected_rc == 2:
                raise RuntimeError(
                    "Connection refused - invalid client identifier"
                )
                break

            elif self.connected_rc == 3:
                raise RuntimeError("Connection refused - server unavailable")
                break

            elif self.connected_rc == 4:
                raise RuntimeError(
                    "Connection refused - bad username or password"
                )
                break

            elif self.connected_rc == 5:
                raise RuntimeError("Connection refused - not authorised")
                break

        topics = list()
        topics.append(["service/commands/firmware/" + self.device.key, 2])
        topics.append(["service/commands/file/" + self.device.key, 2])
        topics.append(["service/commands/url/" + self.device.key, 2])
        topics.append(["service/binary/" + self.device.key, 2])
        topics.append(["pong/" + self.device.key, 2])
        topics.append(["configurations/commands/" + self.device.key, 2])

        if self.device.actuator_references:

            for actuator_reference in self.device.actuator_references:

                topics.append(
                    [
                        "actuators/commands/"
                        + self.device.key
                        + "/"
                        + actuator_reference,
                        2,
                    ]
                )

        self.topics = topics
        self.logger.debug("calling subscribe with topics: %s", topics)
        self.client.subscribe(topics)
        self.logger.debug("connect ended")

    def disconnect(self):
        """Disconnects the device from the WolkAbout IoT Platform."""
        self.logger.debug("disconnect")
        self.client.publish("lastwill/" + self.device.key, "Gone offline")
        self.client.loop_stop()
        self.client.disconnect()

    def publish(self, outbound_message):
        """
        Publish the outbound_message to the WolkAbout IoT Platform.

        :param outbound_message: message to be published
        :type outbound_message: OutboundMessage

        :returns: result
        :rtype: bool
        """
        if outbound_message is None:
            self.logger.warning("No message to publish!")
            return False

        if not self.connected:
            self.logger.warning(
                "Not connected, unable to publish. Channel: %s ; Payload: %s",
                outbound_message.channel,
                outbound_message.payload,
            )
            return False

        info = self.client.publish(
            outbound_message.channel, outbound_message.payload, self.qos
        )

        if info.rc == mqtt.MQTT_ERR_SUCCESS:

            self.logger.debug(
                "Published to channel: %s ; Payload: %s",
                outbound_message.channel,
                outbound_message.payload,
            )
            return True

        elif info.is_published():

            self.logger.debug(
                "Published to channel: %s ; Payload: %s",
                outbound_message.channel,
                outbound_message.payload,
            )
            return True

        else:

            self.logger.warning(
                "Publishing failed! Channel: %s ; Payload: %s",
                outbound_message.channel,
                outbound_message.payload,
            )
            return False



#   Copyright 2018 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
Actuator Status Provider module.

Contains ActuatorStatusProvider "interface".
"""


class ActuatorStatusProvider:
    """
    Must be implemented to provide information about current actuator status.

    Should be implemented as non blocking.
    """

    def get_actuator_status(self, reference):
        """
        Return the current actuator state and value.

        The possible states are:

            wolk.ACTUATOR_STATE_READY
            wolk.ACTUATOR_STATE_BUSY
            wolk.ACTUATOR_STATE_ERROR

        :param reference: reference of the actuator
        :type reference: str

        :returns: (state, value)
        :rtype: (int, int or float or str)
        """
        pass



#   Copyright 2018 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
    ActuatorStatusProvider reads the status of your device's actuators. This information is passed on to the platform.
"""


class ActuatorStatusProvider:
    """
        Actuator Status Provider Interface
    """

    def get_actuator_status(self, reference):
        """
        Reads the status of actuator from the device
        and returns it as a tuple containing the ActuatorState and current value.
        Must be implemented as non blocking.
        Must be implemented as thread safe.
        :param reference:
        :type reference: str
        :returns: (state, value)
        :rtype: (ActuatorState, int/str)
        """
        pass



#   Copyright 2018 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
    Holds all data contained from a firmware chunk request response
"""


class FileTransferPacket:
    """
        File Transfer Packet
    """

    def __init__(self, previous_hash, data, current_hash):
        """
        :param previous_hash: Hash of the previous chunk
        :type previous_hash: bytes
        :param data: The requested chunk
        :type data: bytes
        :param current_hash: Hash of the current chunk
        :type current_hash: bytes
        """
        self.previous_hash = previous_hash
        self.data = data
        self.current_hash = current_hash



#   Copyright 2018 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
    Implement this class to wire and read files in persistent memory, persist firmware version
    and handle the process of installation of firmware itself.
    Optionally provide implementation for URL download
"""


class FirmwareHandler:
    """
        Firmware Handler Interface
    """

    def update_start(self, file_name, file_size):
        """
        Prepares the device to start the process of firmware update
        """
        pass

    def update_finalize(self):
        """
        Finishes the process of firmware update
        """
        pass

    def update_abort(self):
        """
        Aborts the process of firmware update, returning the device to the idle state
        """
        pass

    def write_chunk(self, chunk):
        """
        Writes the firmware file chunk into persitent memory
        """
        pass

    def read_chunk(self, index):
        """
        Reads the firmware file chunk from persistent memory
        """
        pass

    def persist_version(self, version):
        """
        Places the current firmware version into a persistent queue of messages to be sent to the platform
        """
        pass

    def unpersist_version(self):
        """
        Removed the persisted firmwave version from the queue
        """
        pass

    def update_start_url_download(self, file_url):
        """
        Prepares the device for the URL file transfer
        """
        pass

    def set_url_download_result_callback(self, callback):
        """
        Sets the callback method for reporting the result of URL file download
        """
        pass



#   Copyright 2018 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
OS Outbound Message Queue module.

Contains OSOutboundMessageQueue class used to store messages.
"""
from collections import deque

from wolk.wolkcore import OutboundMessageQueue
from wolk import LoggerFactory


class OSOutboundMessageQueue(OutboundMessageQueue):
    """
    Store messages before they are sent to the WolkAbout IoT Platform.

    :ivar logger: Logger instance issued from the LoggerFactory class
    :vartype logger: logger
    :ivar queue: Double ended queue used to store messages
    :vartype queue: collections.deque
    """

    def __init__(self):
        """Create a double ended queue to store messages."""
        self.queue = deque()
        self.logger = LoggerFactory.logger_factory.get_logger(
            str(self.__class__.__name__)
        )
        self.logger.debug("Init")

    def put(self, message):
        """
        Add the message to the queue.

        Combines sensor reading messages of the same reference
        into a single message.

        :param message: message to place in the queue
        :type message: OutboundMessage

        :returns: Nothing
        :rtype: None
        """
        if not message:
            return

        if "readings" not in message.channel:
            self.queue.append(message)
            self.logger.debug(
                "put - Queue size: %s ; Channel: %s ; Payload: %s",
                len(self.queue),
                message.channel,
                message.payload,
            )

        reading_reference = message.channel.split("/")[-1]

        present_in_queue = False

        for stored_message in self.queue:
            if "readings" not in message.channel:
                continue
            if reading_reference == stored_message.channel.split("/")[-1]:
                present_in_queue = True
                break

        if not present_in_queue:
            self.queue.append(message)
            self.logger.debug(
                "put - Queue size: %s ; Channel: %s ; Payload: %s",
                len(self.queue),
                message.channel,
                message.payload,
            )
            return

        readings = 0
        max_data = 1

        for stored_message in self.queue:
            if "readings" not in stored_message.channel:
                continue
            if reading_reference == stored_message.channel.split("/")[-1]:
                readings += 1
                data_count = stored_message.payload.count('"data"')
                if data_count > max_data:
                    max_data = data_count

        if readings > 0 and max_data == 1:
            for stored_message in self.queue:
                if reading_reference == stored_message.channel.split("/")[-1]:
                    stored_message.payload = "[" + stored_message.payload
                    stored_message.payload += "," + message.payload + "]"
                    self.logger.debug(
                        "put - Queue size: %s ; Channel: %s ; Payload: %s",
                        len(self.queue),
                        message.channel,
                        message.payload,
                    )
                    break

        if max_data > 1:
            for stored_message in self.queue:
                if "readings" not in stored_message.channel:
                    continue
                if reading_reference == stored_message.channel.split("/")[-1]:
                    data_count = stored_message.payload.count('"data"')
                    if max_data > data_count:
                        continue

                    stored_message.payload = stored_message.payload[:-1]
                    stored_message.payload += "," + message.payload + "]"
                    self.logger.debug(
                        "put - Queue size: %s ; Channel: %s ; Payload: %s",
                        len(self.queue),
                        message.channel,
                        message.payload,
                    )

    def get(self):
        """
        Take the first message from the queue.

        :returns: message
        :rtype: None if empty queue or OutboundMessage
        """
        if len(self.queue) == 0:
            return None

        message = self.queue.popleft()
        self.logger.debug(
            "get - Queue size: %s ; Channel: %s ; Payload: %s",
            len(self.queue),
            message.channel,
            message.payload,
        )
        return message

    def peek(self):
        """
        Return the first message from the queue without removing it.

        :returns: message
        :rtype: OutboundMessage
        """
        if len(self.queue) == 0:

            self.logger.debug("peek - Queue size: %s", len(self.queue))
            return

        else:
            message = self.queue[0]
            self.logger.debug(
                "peek - Queue size: %s ; Channel: %s ; Payload: %s",
                len(self.queue),
                message.channel,
                message.payload,
            )
            return message



#   Copyright 2018 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


"""
    Provide implementation of ActuationHandler to pass actuation commands from the platform to your device.
"""


class ActuationHandler:
    """
        Actuation Handler Interface
    """

    def handle_actuation(self, reference, value):
        """
        When the actuation command is given from the platform, it will be delivered to this method.
        This method should pass the new value for the actuator to the device.
        Must be implemented as non blocking.
        Must be implemented as thread safe.

        :param reference: The reference of the actuator
        :type reference: str
        :param value: The value to which to set the actuator
        :type value: int, float, string
        """
        pass



#   Copyright 2018 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
    This interface will publish messages to the platform in regular intervals to keep device connected in cases
    where no data is being sent for over 30 minutes
"""


class KeepAliveService:
    """
        Keep Alive Service interface
    """

    def handle_pong():
        """
        Currently unused
        """
        pass

    def start():
        """
        Publish a ping message in a repeating interval
        """
        pass

    def stop():
        """
        Stop the repeating interval
        """
        pass



#   Copyright 2018 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
    Provide implementation of ConnectivityService to enable connecting to the platform
"""


class ConnectivityService:
    """
        Connectivity Service Interface
    """

    def connect(self):
        pass

    def disconnect(self):
        pass

    def connected(self):
        """
        Returns current state of the connection to the platform
        :returns: state
        :rtype: bool
        """
        pass

    def publish(self, outbound_message):
        """
        Publishes message to the platform.
        Returns true on success, false otherwise.
        :param outbound_message: Message to send
        :type outbound_message: OutboundMessage
        :returns: success
        :rtype: bool
        """
        pass

    def set_inbound_message_listener(self, listener):
        """
        Set a callback to Wolk._on_inbound_message method.
        This callback function needs to pass back an InboundMessage object
        :param listener:
        :type listener: Wolk
        :return:
        """
        pass



#   Copyright 2018 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging


class LoggerFactory:
    """Factory for issuing ready to use loggers in other modules."""

    def __init__(
        self,
        level=logging.INFO,
        log_format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        console=True,
        log_file=None,
    ):
        """
        Create a factory that will give loggers through calls to get_logger().

        :param level: Set the desired logging level
        :type level: int, optional
        :param log_format: Desired logging format
        :type log_format: str, optional
        :param console: Should the log messages be outputted to the console
        :type console: bool, optional
        :param log_file: Name of the log file to output to
        :type log_file: None, optional
        """
        self.level = level
        self.log_format = log_format
        self.console = console
        self.log_file = log_file

    def get_logger(self, name, level=None):
        """Return a ready to use logger instance.

        :param name: Name of the logger
        :type name: str
        :param level: Override the log level
        :type level: None, optional

        :returns: Logger instance
        :rtype: logger
        """
        logger = logging.getLogger(name)

        if level is not None:
            logger.setLevel(level)
        else:
            logger.setLevel(self.level)

        formatter = logging.Formatter(self.log_format)

        if self.console:

            console_handler = logging.StreamHandler()
            if level is not None:
                console_handler.setLevel(level)
            else:
                console_handler.setLevel(self.level)

            console_handler.setFormatter(formatter)
            logger.addHandler(console_handler)

        if self.log_file is not None:
            file_handler = logging.FileHandler(self.log_file)
            if level is not None:
                file_handler.setLevel(level)
            else:
                file_handler.setLevel(self.level)
            file_handler.setFormatter(formatter)
            logger.addHandler(file_handler)

        return logger


# Logging levels available: NOTSET, INFO, DEBUG
logger_factory = LoggerFactory(level=logging.INFO)


def logging_config(level, log_file=None):
    """
    Set desired log level and designate a log file.

    :param level: Available levels : debug, info, notset
    :type level: str
    :param log_file: path to log file
    :type log_file: str or None, optional
    """
    if level == "debug":
        logger_factory.level = logging.DEBUG
    elif level == "info":
        logger_factory.level = logging.INFO
    elif level == "notset":
        logger_factory.level = logging.NOTSET

    if log_file is not None:
        logger_factory.log_file = log_file



#   Copyright 2018 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
    Holds all the necessary information to communicate with the platform
"""


class Device:
    """
        Device class
    """

    def __init__(self, key, password, actuator_references=None):
        """
        :param key: Username used to connect to the platform
        :type key: str
        :param password: Password used to authenticate the connection
        :type password: str
        :param actuator_references: List of actuator references defined in the device manifest
        :type actuator_references: list
        """
        self.key = key
        self.password = password
        self.actuator_references = actuator_references



# coding=utf-8
"""
This module provides connection to WolkAbout IoT Platform.

To start publishing data to the platform
create an instance of Device class with credentials obtained from the platform
and pass it to an instance of WolkConnect class.

For more information about module features visit:
https://github.com/Wolkabout/WolkConnect-Python/tree/master/examples/full_feature_set


Throughout this package usage of enumerations and ABCs are omitted
due to a constraint caused by the necessity to have
a single base (wolkcore) for two versions of python: CPython and Zerynth.

:ivar ACTUATOR_STATE_BUSY: Actuator state busy
:vartype ACTUATOR_STATE_BUSY: int
:ivar ACTUATOR_STATE_ERROR: Actuator state error
:vartype ACTUATOR_STATE_ERROR: int
:ivar ACTUATOR_STATE_READY: Actuator state ready
:vartype ACTUATOR_STATE_READY: int
:ivar VERSION_MAJOR: Major version
:vartype VERSION_MAJOR: int
:ivar VERSION_MINOR: Minor version
:vartype VERSION_MINOR: int
:ivar VERSION_PATCH: Patch version
:vartype VERSION_PATCH: int
"""
from .ActuationHandler import ActuationHandler
from .ActuatorStatusProvider import ActuatorStatusProvider
from .ConfigurationHandler import ConfigurationHandler
from .ConfigurationProvider import ConfigurationProvider
from .Device import Device
from .FileSystemFirmwareHandler import FileSystemFirmwareHandler
from .FirmwareInstaller import FirmwareInstaller
from .FirmwareURLDownloadHandler import FirmwareURLDownloadHandler
from .LoggerFactory import logging_config
from .wolkcore.OutboundMessageQueue import OutboundMessageQueue
from .WolkConnect import WolkConnect


__all__ = [
    "ActuationHandler",
    "ActuatorStatusProvider",
    "ConfigurationHandler",
    "ConfigurationProvider",
    "Device",
    "FileSystemFirmwareHandler",
    "FirmwareInstaller",
    "FirmwareURLDownloadHandler",
    "LoggerFactory",
    "logging_config",
    "OutboundMessageQueue",
    "WolkConnect",
]

# "Enum" of actuator states
ACTUATOR_STATE_READY = 0
ACTUATOR_STATE_BUSY = 1
ACTUATOR_STATE_ERROR = 2


# "Enum" of connector version
VERSION_MAJOR = 3
VERSION_MINOR = 0
VERSION_PATCH = 4



#   Copyright 2018 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
OS Inbound Message Deserializer module.

Contains OSInboundMessageDeserializer class.
"""
import json

from wolk.wolkcore import ActuatorCommand
from wolk.wolkcore import ActuatorCommandType
from wolk.wolkcore import ConfigurationCommand
from wolk.wolkcore import ConfigurationCommandType
from wolk.wolkcore import FileTransferPacket
from wolk.wolkcore import FirmwareCommand
from wolk.wolkcore import FirmwareCommandType
from wolk.wolkcore import InboundMessageDeserializer
from wolk import LoggerFactory


class OSInboundMessageDeserializer(InboundMessageDeserializer):
    """
    Deserializes messages received from the WolkAbout IoT Platform.

    :ivar logger: Logger instance issued from the LoggerFactory class
    :vartype logger: logger
    """

    def __init__(self):
        """Just logger init."""
        self.logger = LoggerFactory.logger_factory.get_logger(
            str(self.__class__.__name__)
        )
        self.logger.debug("Init")

    def deserialize_actuator_command(self, message):
        """
        Deserialize the message into an actuation command.

        :param message: message to be deserialized
        :type message: InboundMessage

        :returns: actuation
        :rtype: ActuatorCommand
        """
        self.logger.debug("deserialize_actuator_command called")
        reference = message.channel.split("/")[-1]
        payload = json.loads(message.payload)
        command = payload.get("command")

        if str(command) == "SET":

            command_type = ActuatorCommandType.ACTUATOR_COMMAND_TYPE_SET
            value = payload.get("value")
            if value == "true":
                value = True
            elif value == "false":
                value = False

            actuation = ActuatorCommand(reference, command_type, value)
            self.logger.info(
                "Received actuation command - Reference: %s ; Command: SET ;"
                " Value: %s",
                actuation.reference,
                actuation.value,
            )
            return actuation

        elif str(command) == "STATUS":

            command_type = ActuatorCommandType.ACTUATOR_COMMAND_TYPE_STATUS
            actuation = ActuatorCommand(reference, command_type)
            self.logger.info(
                "Received actuation command - Reference: %s ;"
                " Command: STATUS ",
                actuation.reference,
            )
            return actuation

        else:

            command_type = ActuatorCommandType.ACTUATOR_COMMAND_TYPE_UNKNOWN
            actuation = ActuatorCommand(reference, command_type)
            self.logger.warning(
                "Received actuation command - Reference: %s ;"
                " Command: %s (UNKNOWN)",
                actuation.reference,
                actuation.command,
            )
            return actuation

    def deserialize_firmware_command(self, message):
        """
        Deserialize the message into a firmware command.

        :param message: message to be deserialized
        :type message: InboundMessage

        :returns: firmware_command
        :rtype: FirmwareCommand
        """
        self.logger.debug("deserialize_firmware_command called")
        payload = json.loads(message.payload)
        command = payload.get("command")

        if command == "FILE_UPLOAD":

            command = FirmwareCommandType.FIRMWARE_COMMAND_TYPE_FILE_UPLOAD
            firmware_command = FirmwareCommand(
                command,
                payload.get("fileName"),
                payload.get("fileSize"),
                payload.get("fileHash"),
                payload.get("autoInstall"),
            )
            self.logger.debug(
                "deserialize_firmware_command - Command: %s (FILE_UPLOAD) ; "
                "File name: %s ; File size: %s ; "
                "File hash: %s ; Auto install: %s",
                firmware_command.command,
                firmware_command.file_name,
                firmware_command.file_size,
                firmware_command.file_hash,
                firmware_command.auto_install,
            )
            return firmware_command

        elif command == "URL_DOWNLOAD":

            command = FirmwareCommandType.FIRMWARE_COMMAND_TYPE_URL_DOWNLOAD
            firmware_command = FirmwareCommand(
                command,
                file_url=payload.get("fileUrl"),
                auto_install=payload.get("autoInstall"),
            )
            self.logger.debug(
                "deserialize_firmware_command - Command: %s (URL_DOWNLOAD) ; "
                "File url: %s ; Auto install: %s",
                firmware_command.command,
                firmware_command.file_url,
                firmware_command.auto_install,
            )
            return firmware_command

        elif command == "INSTALL":

            command = FirmwareCommandType.FIRMWARE_COMMAND_TYPE_INSTALL

            firmware_command = FirmwareCommand(command)
            self.logger.debug(
                "deserialize_firmware_command - Command: %s (INSTALL)",
                firmware_command.command,
            )
            return firmware_command

        elif command == "ABORT":

            command = FirmwareCommandType.FIRMWARE_COMMAND_TYPE_ABORT

            firmware_command = FirmwareCommand(command)
            self.logger.debug(
                "deserialize_firmware_command - Command: %s (ABORT)",
                firmware_command.command,
            )
            return firmware_command

        else:

            command = FirmwareCommandType.FIRMWARE_COMMAND_TYPE_UNKNOWN

            firmware_command = FirmwareCommand(command)
            self.logger.debug(
                "deserialize_firmware_command - Command: %s (UNKNOWN)",
                firmware_command.command,
            )
            return firmware_command

    def deserialize_firmware_chunk(self, message):
        """
        Split the message into a packet.

        :param message: message to be deserialized
        :type message: InboundMessage

        :returns: packet
        :rtype: FileTransferPacket
        """
        self.logger.debug("deserialize_firmware_chunk called")
        previous_hash = message.payload[:32]
        data = message.payload[32 : len(message.payload) - 32]
        current_hash = message.payload[-32:]

        packet = FileTransferPacket(previous_hash, data, current_hash)
        self.logger.debug(
            "deserialize_firmware_chunk - Previous hash: %s ; "
            "Data size: %s ; Current hash: %s",
            packet.previous_hash,
            len(packet.data),
            packet.current_hash,
        )
        return packet

    def deserialize_configuration_command(self, message):
        """
        Deserialize the message into a configuration command.

        :param message: message to be deserialized
        :type message: InboundMessage

        :returns: configuration
        :rtype: ConfigurationCommand
        """
        self.logger.debug("deserialize_configuration_command called")
        payload = json.loads(message.payload)
        command = payload.get("command")

        if command == "SET":

            command = ConfigurationCommandType.CONFIGURATION_COMMAND_TYPE_SET

            configuration = ConfigurationCommand(
                command, payload.get("values")
            )

            self.logger.info(
                "Received configuration command - Command: SET  Values: %s",
                configuration.values,
            )

            for (
                received_reference,
                received_value,
            ) in configuration.values.items():
                if "," in received_value:
                    values_list = received_value.split(",")
                    try:
                        if any("." in value for value in values_list):
                            values_list = [
                                float(value) for value in values_list
                            ]
                        else:
                            values_list = [int(value) for value in values_list]
                    except ValueError:
                        pass

                        configuration.values[received_reference] = tuple(
                            values_list
                        )
                        break

            return configuration

        elif command == "CURRENT":

            command = (
                ConfigurationCommandType.CONFIGURATION_COMMAND_TYPE_CURRENT
            )

            configuration = ConfigurationCommand(command)
            self.logger.info(
                "Received configuration command - Command: CURRENT"
            )
            return configuration

        else:

            command = (
                ConfigurationCommandType.CONFIGURATION_COMMAND_TYPE_UNKNOWN
            )

            configuration = ConfigurationCommand(command)
            self.logger.warning(
                "Received configuration command - Command: %s (UNKONWN)",
                configuration.command,
            )
            return configuration



#   Copyright 2018 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
OS Keep Alive Service module.

Contains OSKeepAliveService class used to send ping messages to the platform.
Contains RepeatingTimer class used to keep sending pings in an interval.
"""

from wolk import LoggerFactory
from wolk.wolkcore import KeepAliveService
from threading import Timer


class OSKeepAliveService(KeepAliveService):
    """
    Send messages to platform in regular intervals to keep device connected.

    Used for cases where no data is being sent by device for over 30 minutes.

    :ivar connectivity_service: Connectivity service used to publish
    :vartype connectivity_service: ConnectivityService
    :ivar interval: number of seconds between each keep alive message
    :vartype interval: int
    :ivar logger: Logger instance issued from the LoggerFactory class
    :vartype logger: logger
    :ivar outbound_message_factory: message factory used to create pings
    :vartype outbound_message_factory: OutboundMessageFactory
    :ivar timer: timer for sending ping messages
    :vartype timer: RepeatingTimer
    """

    def __init__(
        self, connectivity_service, outbound_message_factory, interval=600
    ):
        """
        Service for sending keep alive messages.

        :param connectivity_service: Connectivity service used to publish
        :type connectivity_service: ConnectivityService
        :param outbound_message_factory: message factory used to create pings
        :type outbound_message_factory: OutboundMessageFactory
        :param interval: number of seconds between each keep alive message
        :type interval: int, optional
        """
        self.logger = LoggerFactory.logger_factory.get_logger(
            str(self.__class__.__name__)
        )
        self.logger.debug(
            "Init:  Connectivity service: %s ; Outbound message factory: %s "
            "Interval: %s",
            connectivity_service,
            outbound_message_factory,
            interval,
        )
        self.connectivity_service = connectivity_service
        self.outbound_message_factory = outbound_message_factory
        self.interval = interval
        self.timer = None

    def handle_pong(self):
        """Handle keep alive response message received from the platform."""
        self.logger.debug("handle_pong called")
        pass

    def start(self):
        """
        Send a keep alive message as soon as the device is connected.

        Starts a repeating timer to send subsequent keep alive messages
        every self.interval seconds
        """
        self.logger.debug("start called")

        # Timer to give connectivity service time to realize it is connected
        t = Timer(5, self.send_keep_alive)
        t.start()

        # First message will fail to send
        # due to connectivity service not realizing it is connected
        self.timer = RepeatingTimer(self.interval, self.send_keep_alive)
        self.timer.start()

    def stop(self):
        """Stop the repeating timer."""
        self.logger.debug("stop called")
        self.timer.cancel()

    def send_keep_alive(self):
        """
        Make ping and send.

        Create a keep alive message from the outbound message factory
        and publishes it using the connectivity service
        """
        self.logger.debug("send_keep_alive called")
        message = self.outbound_message_factory.make_from_keep_alive_message()
        self.connectivity_service.publish(message)


class RepeatingTimer:
    """
    Repeating timer calls function with arguments every interval.

    :ivar args: list of arguments
    :vartype args: list
    :ivar f: function to call
    :vartype f: method
    :ivar interval: number of seconds after which to call function
    :vartype interval: int
    :ivar kwargs: keyword arguments
    :vartype kwargs: kwargs
    :ivar timer: timer
    :vartype timer: Timer
    """

    def __init__(self, interval, f, *args, **kwargs):
        """
        Create a repeating timer.

        :param interval: number of seconds after which to call function
        :type interval: int
        :param f: function to call
        :type f: method
        :param *args: list of arguments
        :type *args: list
        :param **kwargs: keyword arguments
        :type **kwargs: kwargs
        """
        self.interval = interval
        self.f = f
        self.args = args
        self.kwargs = kwargs

        self.timer = None

    def callback(self):
        """Call function and start timer."""
        self.f(*self.args, **self.kwargs)
        self.start()

    def cancel(self):
        """Cancel timer."""
        self.timer.cancel()

    def start(self):
        """Start timer."""
        self.timer = Timer(self.interval, self.callback)
        self.timer.start()



#   Copyright 2018 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
    Responsible for handling file and URL download, as well as reporting firmware status.
    Should be implemented to depend upon an object of FirmwareHandler class
"""


class FirmwareUpdate:
    """
    Firmware Update class
    """

    def handle_file_download(self, firmware_command):
        """
        Passes the information necessary to start a file download to an instance of FirmwareHandler
        """
        pass

    def handle_url_download(self, firmware_command):
        """
        Passes the information necessary to start a URL download to an instance of FirmwareHandler
        """
        pass

    def handle_url_download_result(self, result):
        """
        Handles the result of the url file download reported from an instance of FirmwareHandler
        """
        pass

    def handle_install(self):
        """
        Passes the install command to an instance of FirmwareHandler
        """
        pass

    def handle_abort(self):
        """
        Passes the abort command to an instance of FirmwareHandler
        """
        pass

    def handle_packet(self, packet):
        """
        Pases the packet to an instance of FirmwareHandler
        """
        pass

    def report_result(self):
        """
        Reports the results of the firmware update process
        """
        pass

    def set_on_file_packet_request_callback(self, on_file_packet_request_callback):
        """
        Sets a callback to handle file packet requests
        """
        pass

    def set_on_status_callback(self, on_status_callback):
        """
        Sets a callback to handle firmware status reporting
        """
        pass



#   Copyright 2018 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
Firmware Installer module.

Contains FirmwareInstaller "interface" that handles firmware installation.
"""


class FirmwareInstaller:
    """Implement to handle installation of the firmware file itself."""

    def install_firmware(self, firmware_file_path):
        """
        Handle the installation of the firmware file.

        :param firmware_file_path: path where the firmware file is located
        :type firmware_file_path: str
        """
        pass
